// Physical memory allocator, for user processes,
// kernel stacks, page-table pages,
// and pipe buffers. Allocates whole 4096-byte pages.

#include "include/types.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/riscv.h"
#include "include/spinlock.h"
#include "include/kalloc.h"
#include "include/string.h"
#include "include/printf.h"

void freerange(void *pa_start, void *pa_end);

extern char kernel_end[]; // first address after kernel.

struct run
{
  struct run *next;
};

#define MAX_ORDER 11 // 支持到2^11=2048页
#define MIN_ORDER 0  // 最小分配单位为1页

struct free_area
{
  struct run *free_list;
  uint64 nr_free;
};

struct
{
  struct spinlock lock;
  struct free_area free_area[MAX_ORDER + 1];
} buddy;

// 初始化伙伴系统
void kinit()
{
  initlock(&buddy.lock, "buddy");
  for (int i = 0; i <= MAX_ORDER; i++)
  {
    buddy.free_area[i].free_list = 0;
    buddy.free_area[i].nr_free = 0;
  }
  // 初始化空闲内存
  freerange(kernel_end, (void *)PHYSTOP);
}

// 计算对数
static inline int
get_order(uint64 size)
{
  int order = 0;
  size = size >> PGSHIFT;
  while (size > 1)
  {
    size >>= 1;
    order++;
  }
  return order;
}

// 分配2^order个页面
void *
buddy_alloc(int order)
{
  struct run *block;
  int current_order = order;

  acquire(&buddy.lock);

  // 查找可用的大块内存
  for (; current_order <= MAX_ORDER; current_order++)
  {
    if (!buddy.free_area[current_order].nr_free)
      continue;

    block = buddy.free_area[current_order].free_list;
    buddy.free_area[current_order].free_list = block->next;
    buddy.free_area[current_order].nr_free--;

    // 分裂大块
    while (current_order > order)
    {
      current_order--;
      struct run *buddy_block = (struct run *)((char *)block + (1 << (current_order + PGSHIFT)));
      buddy_block->next = buddy.free_area[current_order].free_list;
      buddy.free_area[current_order].free_list = buddy_block;
      buddy.free_area[current_order].nr_free++;
    }

    release(&buddy.lock);
    return (void *)block;
  }

  release(&buddy.lock);
  return 0;
}

// 替换原有的kalloc
void *
kalloc(void)
{
  // 计算PGSIZE对应的order
  int order = get_order(PGSIZE);
  void *p = buddy_alloc(order);
  if (p)
    memset(p, 5, PGSIZE);
  return p;
}

// 释放内存块
void kfree(void *p)
{
  if (((uint64)p % PGSIZE) != 0 || p < (void *)kernel_end || (uint64)p >= PHYSTOP)
    panic("kfree");

  acquire(&buddy.lock);

  // 使用get_order计算块大小
  int order = get_order(PGSIZE);
  struct run *block = (struct run *)p;

  while (order < MAX_ORDER)
  {
    uint64 buddy_addr = (uint64)p ^ (1 << (order + PGSHIFT));
    struct run *buddy_block = (struct run *)buddy_addr;

    // 检查伙伴块是否空闲
    int found = 0;
    struct run **prev = &buddy.free_area[order].free_list;
    for (struct run *r = *prev; r; prev = &r->next, r = r->next)
    {
      if (r == buddy_block)
      {
        *prev = r->next;
        buddy.free_area[order].nr_free--;
        found = 1;
        break;
      }
    }

    if (!found)
      break;

    // 合并伙伴块
    if (buddy_addr < (uint64)block)
      block = buddy_block;
    order++;
  }

  // 将合并后的块加入空闲列表
  block->next = buddy.free_area[order].free_list;
  buddy.free_area[order].free_list = block;
  buddy.free_area[order].nr_free++;

  release(&buddy.lock);
}

void freerange(void *pa_start, void *pa_end)
{
  char *p;
  p = (char *)PGROUNDUP((uint64)pa_start);
  for (; p + PGSIZE <= (char *)pa_end; p += PGSIZE)
    kfree(p);
}

uint64
freemem_amount(void)
{
  return buddy.free_area[0].nr_free << PGSHIFT;
}
