#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/sleeplock.h"
#include "include/file.h"
#include "include/pipe.h"
#include "include/kalloc.h"
#include "include/vm.h"

pipe_status_t
pipealloc(struct file **f0, struct file **f1)
{
    struct pipe *pi;

    if (!f0 || !f1)
        return PIPE_INVALID_PARAM;

    pi = 0;
    *f0 = *f1 = 0;
    if((*f0 = filealloc()) == NULL || (*f1 = filealloc()) == NULL)
        goto bad;
    if((pi = (struct pipe*)kalloc()) == NULL)
        goto bad;

    pi->readopen = 1;
    pi->writeopen = 1;  
    pi->nwrite = 0;
    pi->nread = 0;
    pi->error = 0;
    // pi->last_op = ticks;
    
    initlock(&pi->lock, "pipe");
    (*f0)->type = FD_PIPE;
    (*f0)->readable = 1;
    (*f0)->writable = 0;
    (*f0)->pipe = pi;
    (*f1)->type = FD_PIPE;
    (*f1)->readable = 0;
    (*f1)->writable = 1;
    (*f1)->pipe = pi;

    return PIPE_OK;

bad:
    if(pi)
        kfree((char*)pi);
    if(*f0)
        fileclose(*f0);
    if(*f1)
        fileclose(*f1);
    return PIPE_ERROR;
}

pipe_status_t
pipeclose(struct pipe *pi, int writable)
{
    // Parameter validation
    if (!pi) {
        return PIPE_INVALID_PARAM;
    }

    acquire(&pi->lock);

    // Update pipe state
    if (writable) {
        pi->writeopen = 0;
        wakeup(&pi->nread);
    } else {
        pi->readopen = 0;
        wakeup(&pi->nwrite);
    }

    // Check if pipe can be freed
    if (pi->readopen == 0 && pi->writeopen == 0) {
        release(&pi->lock);
        kfree((char*)pi);
        return PIPE_CLOSED;
    }

    release(&pi->lock);
    return PIPE_OK;
}

pipe_status_t
pipewrite(struct pipe *pi, uint64 addr, int n)
{
    // Parameter validation
    if (!pi || !addr || n < 0) {
        return PIPE_INVALID_PARAM;
    }

    int i;
    char ch;
    struct proc *pr = myproc();
    // uint64 start_time = ticks;

    acquire(&pi->lock);
    for(i = 0; i < n; i++) {
        // Check for full pipe
        while(pi->nwrite == pi->nread + PIPESIZE) {
            // Check for closed pipe or killed process
            if(pi->readopen == 0) {
                release(&pi->lock);
                return PIPE_CLOSED;
            }
            if(pr->killed) {
                release(&pi->lock);
                return PIPE_ERROR;
            }
            // // Check for timeout
            // if(ticks - start_time > PIPE_TIMEOUT_TICKS) {
            //     release(&pi->lock);
            //     return PIPE_TIMEOUT;
            // }
            
            wakeup(&pi->nread);
            sleep(&pi->nwrite, &pi->lock);
        }

        // Copy data from user space
        if(copyin2(&ch, addr + i, 1) == -1) {
            release(&pi->lock);
            return PIPE_ERROR;
        }

        pi->data[pi->nwrite++ % PIPESIZE] = ch;
        // pi->last_op = ticks;
    }

    wakeup(&pi->nread);
    release(&pi->lock);
    return i > 0 ? PIPE_OK : PIPE_ERROR;
}

pipe_status_t
piperead(struct pipe *pi, uint64 addr, int n)
{
    // Validate parameters
    if (!pi || !addr || n < 0) {
        return PIPE_INVALID_PARAM;
    }

    int i;
    struct proc *pr = myproc();
    char ch;
    // uint64 start_time = ticks;

    acquire(&pi->lock);
    while(pi->nread == pi->nwrite && pi->writeopen) {
        if(pr->killed) {
            release(&pi->lock);
            return PIPE_ERROR;
        }
        // if(ticks - start_time > PIPE_TIMEOUT_TICKS) {
        //     release(&pi->lock);
        //     return PIPE_TIMEOUT;
        // }
        sleep(&pi->nread, &pi->lock);
    }

    // Handle closed pipe with no data
    if(pi->nread == pi->nwrite && !pi->writeopen) {
        release(&pi->lock);
        return PIPE_CLOSED;
    }

    for(i = 0; i < n; i++) {
        if(pi->nread == pi->nwrite)
            break;
        ch = pi->data[pi->nread++ % PIPESIZE];
        if(copyout2(addr + i, &ch, 1) == -1) {
            release(&pi->lock);
            return PIPE_ERROR;
        }
        // pi->last_op = ticks;
    }

    wakeup(&pi->nwrite);
    release(&pi->lock);
    return i > 0 ? PIPE_OK : PIPE_ERROR;
}